module ngeq where

open import Data.Unit
open import Data.Product
open import Data.Nat
open import Data.Sum
open import Data.List hiding ([_]; all)
open import Data.Vec hiding ([_]; _⊛_)
open import Data.Vec1
open import Data.Bool

open import local-lib
open import nvec-map
open import data-view

toBool : {n : ℕ} → (v : Vec₁ Set n) → Set
toBool {zero}   []         = Bool
toBool {suc n}  (A1 ∷ As)  = A1 → toBool As

NGeq : {n : ℕ} → (v : Vec₁ Set n) → Set1
NGeq v = ⇑ (toBool v)

defUnit : (n : ℕ) → NGeq (repeat₁ ⊤)
defUnit zero     = ↑ \ x → true
defUnit (suc n)  = ↑ \ x → ↓ (defUnit n)

constFalse : {n : ℕ} → (v : Vec₁ Set n) → toBool v
constFalse {zero}  []   = false
constFalse {suc m} (A1 ∷ As)  = \ a → constFalse As

defNat : (n : ℕ) → NGeq (repeat₁ {n} ℕ)
defNat zero     = ↑ true
defNat (suc zero)  = ↑ \ x → true
defNat (suc (suc n))  = 
  ↑ \ x → \ y → if eqNat x y  then ↓ (defNat (suc n)) y
                              else constFalse (repeat₁ ℕ)

defPair : (n : ℕ) → 
        (as : Vec₁ Set n) → (NGeq as) →
        (bs : Vec₁ Set n) → (NGeq bs) → 
        NGeq ((repeat₁ _×_ ⊛₁ as) ⊛₁ bs)
defPair zero     ([])  (↑ at)  ([])  (↑ bt) = ↑ (at ∧ bt)
defPair (suc n)  (a ∷ as)  (↑ af)  (b ∷ bs) (↑ bf) =
  ↑ \ x →
      ↓ (defPair n as (↑ (af (proj₁ x)))
                   bs (↑ (bf (proj₂ x))))

defSumLeft : (n : ℕ) → 
        (as : Vec₁ Set n) → (NGeq as) →
        (bs : Vec₁ Set n) → 
        NGeq (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
defSumLeft zero ([]) at ([])   = at
defSumLeft (suc n) (a ∷ as) (↑ af) (b ∷ bs) = ↑ f 
  where f : a ⊎ b → toBool (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
        f (inj₁ x1) = ↓ (defSumLeft n as (↑ (af x1)) bs)
        f (inj₂ x1) = constFalse (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)

defSumRight : (n : ℕ) → 
        (as : Vec₁ Set (n)) → 
        (bs : Vec₁ Set (n)) → (NGeq bs) →
        NGeq (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
defSumRight zero [] [] bt  = bt 
defSumRight (suc n) (a ∷ as)  (b ∷ bs) (↑ bf) = ↑ f 
  where f : a ⊎ b → toBool (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
        f (inj₁ x1) = constFalse (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
        f (inj₂ x1) = ↓ (defSumRight n as bs (↑ (bf x1)))

defSum : (n : ℕ) → 
        (as : Vec₁ Set (n)) → (NGeq as) →
        (bs : Vec₁ Set (n)) → (NGeq bs) → 
        NGeq (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
defSum zero ([]) at ([]) bt = bt
defSum (suc n) (a ∷ as) (↑ af) (b ∷ bs) (↑ bf) = ↑ f
  where f : a ⊎ b → toBool (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
        f (inj₁ x1) = ↓ (defSumLeft n as (↑ (af x1)) bs)
        f (inj₂ x1) = ↓ (defSumRight n as bs (↑ (bf x1)))

ngeq-const : {n : ℕ} → TyConstEnv NGeq
ngeq-const {n} Nat = defNat (suc n)
ngeq-const {n} Unit = defUnit n
ngeq-const {n} Prod = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (defPair (suc n))
ngeq-const {n} Sum  = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (defSum (suc n))

ngeq-data : {n : ℕ} → DataGen (NGeq {suc n})
ngeq-data {0}      dt (e ∷ [])  (↑ bt) =
  ↑ \ s → bt (DT-from dt s)
ngeq-data {suc n}  dt (e ∷ es)  (↑ bt) = 
  ↑ \ s → ↓ (ngeq-data dt es (↑ (bt (DT-from dt s))))

ngeq : (n : ℕ) → {k : Kind} → (e : Ty k) → 
  NGeq ⟨ k ⟩ (repeat₁ {suc n} (⌊ e ⌋))
ngeq n e = ngen e ngeq-const ngeq-data

nat-eq3 : ℕ → ℕ → ℕ → Bool
nat-eq3 = ↓ (ngeq 2 (Con Nat)) 

t1 = nat-eq3 2 2 2
t2 = nat-eq3 2 1 1

pair-eq1 : {A1 B1 A2 B2 : Set} 
         → (A1 → B1 → Bool) → (A2 → B2 → Bool)
         → (A1 × A2) → (B1 × B2) → Bool
pair-eq1 eq1 eq2 = ↓ (ngeq 1 (Con Prod) (↑ eq1) (↑ eq2))

