{-# OPTIONS --type-in-type #-}
module ngmap-data where

open import Data.Maybe hiding (maybe)
open import Data.Unit
open import Data.Unit1
open import Data.Product
open import Data.Nat
open import Data.Sum
open import Data.List hiding ([_];fromMaybe)
open import Data.Vec  hiding ([_];fromList;toList)
open import Data.Vec1
open import Data.Bool
open import Relation.Binary.PropositionalEquality1

open import local-lib
open import data-view

arrTy : {n : ℕ} → Vec₁ Set (suc n) → Set
arrTy {0}      ( A ∷ [] )   = A
arrTy {suc n}  ( A1 ∷ As )  = A1 → arrTy As

NGmap : {n : ℕ} → Vec₁ Set (suc n) → Set1
NGmap = arrTy

postulate error : (A : Set) → A

mapNat : (n : ℕ) → NGmap ⟨ ⋆ ⟩ (repeat₁ {suc n} ℕ)
mapNat zero  = zero        -- arbitrary
mapNat (suc zero)     = \ x → x         -- return what was given
mapNat (suc (suc n))  = 
  \ x → \ y → if eqNat x y  then mapNat (suc n) y
                            else error _

mapUnit : (n : ℕ) → NGmap {n} ⟨ ⋆ ⟩ (repeat₁ ⊤)
mapUnit zero = tt
mapUnit (suc n)  = \ x → (mapUnit n)

mapPairAux  : (n : ℕ) 
         → (As : Vec₁ Set (suc n)) → arrTy As
         → (Bs : Vec₁ Set (suc n)) → arrTy Bs 
         → arrTy (repeat₁ {suc n} _×_ ⊛₁ As ⊛₁ Bs)
mapPairAux zero     (A ∷ [])   a  (B ∷ [])   b  = (a , b)
mapPairAux (suc n)  (A1 ∷ As)  a  (B1 ∷ Bs)  b  =
 \ p →
  (mapPairAux n As (a (proj₁ p)) Bs (b (proj₂ p)))

mapSumLeft : (n : ℕ) → 
        (as : Vec₁ Set (suc n)) → (NGmap as) →
        (bs : Vec₁ Set (suc n)) → 
        arrTy (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
mapSumLeft zero (a ∷ []) at (b ∷ [])     = inj₁ (at)
mapSumLeft (suc n) (a ∷ as) af (b ∷ bs)  = f 
  where f : a ⊎ b → arrTy (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
        f (inj₁ x1) = mapSumLeft n as (af x1) bs 
        f (inj₂ x2) = mapSumLeft n as (af (error a)) bs

mapSumRight : (n : ℕ) → 
        (as : Vec₁ Set (suc n)) → 
        (bs : Vec₁ Set (suc n)) → (NGmap bs) →
        arrTy (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
mapSumRight zero (a ∷ []) (b ∷ []) bt     = inj₂ bt
mapSumRight (suc n) (a ∷ as) (b ∷ bs) bf  = f 
  where f : a ⊎ b → arrTy (repeat₁ _⊎_ ⊛₁ as ⊛₁ bs)
        f (inj₁ x1) = mapSumRight n as bs (bf (error b))
        f (inj₂ x1) = mapSumRight n as bs (bf x1)

mapSumAux : (n : ℕ) 
         → (As : Vec₁ Set (suc n)) → NGmap As
         → (Bs : Vec₁ Set (suc n)) → NGmap Bs 
         → NGmap (repeat₁ {suc n} _⊎_ ⊛₁ As ⊛₁ Bs)
mapSumAux zero (A ∷ []) a (B ∷ []) b = 
  inj₂ b    --arbitrary
mapSumAux (suc n) (A1 ∷ As) a (B1 ∷ Bs) b = f
  where 
    f : A1 ⊎ B1 → arrTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
    f (inj₁ a1)  = mapSumLeft  n As (a a1) Bs
    f (inj₂ b1)  = mapSumRight n As Bs (b b1)

ngmap-const : {n : ℕ} → TyConstEnv {n} NGmap 
ngmap-const {n} Nat  = mapNat n
ngmap-const {n} Unit = mapUnit n
ngmap-const {n} Prod = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (mapPairAux n)
ngmap-const {n} Sum  = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (mapSumAux n)

ngmap-data : {n : ℕ} → DataGen (NGmap {n})
ngmap-data {0}      dt  (e ∷ [])   bt  = DT-to dt bt
ngmap-data {suc n}  dt  (e1 ∷ es)  bt  =
  \ x → ngmap-data  {n} dt es (bt (DT-from dt x))

ngmap : (n : ℕ) → {k : Kind} → (e : Typ [] k) → 
  NGmap {n} ⟨ k ⟩ (repeat₁ (interp e []))
ngmap n e = ngen e ngmap-const ngmap-data

maybe-map1  : {A B : Set} → (A → B) 
            → Maybe A → Maybe B
maybe-map1  = ngmap 1 maybe 

vec-map0    : {A : Set} {n : ℕ} → A → Vec A n
vec-map0    = ngmap 0 vec 

vec-map1    : {A B : Set} {n : ℕ} 
            → (A → B) → Vec A n → Vec B n
vec-map1    = ngmap 1 vec 

